
#ifndef CDR_OARCHIVE_HPP
#define CDR_OARCHIVE_HPP

#include <iostream>

#include "ordered_oarchive.hpp"
#include "natural_alignment_distance.hpp"

#include "cdr_types.hpp"

namespace boost {
namespace archive {

using marshall::byte_ordering_t;

template<byte_ordering_t Order>
struct cdr_oarchive : 
    public ordered_oarchive_impl<cdr_oarchive<Order>, Order>
{
    cdr_oarchive(std::ostream& os) :
        ordered_oarchive_impl<cdr_oarchive<Order>, Order>(os) {}

    // All elements are pre-aligned to their natural alignment
    template<typename T>
    void pre_align(const T&) 
    { 
        // 'Natural' alignment
        size_t advance_bytes = natural_alignment_distance<sizeof(T)>(os.tellp());

        if (advance_bytes)
        {
            static char buffer[sizeof(T)] = { 0 };
#ifdef INITIALISE_WITH_RECOGNISABLE_VALUE
            static bool initialised = false;

            if (initialised == false)
            {
              std::fill(buffer, buffer + sizeof(buffer), 0xda);
              initialised = true;
            }
#endif

            os.write(buffer, advance_bytes);
        }
    } 

    // Default implementation same as for the ordered binary archive
    template<typename T>
    void save(const T& t, 
              typename boost::enable_if<boost::mpl::contains<cdr_types, T>, T>::type* dummy = 0)
    {
        ordered_oarchive_impl<cdr_oarchive<Order>, Order>::save(t);
    }

    void save(const std::string& s)
    {
        // Prepend the string with the 32-bit length, including the terminator
        uint32_t length = s.size() + 1;
        save(length);

        ordered_oarchive_impl<cdr_oarchive<Order>, Order>::save(s);

        std::string::value_type terminator(0);
        save(terminator);
    }

    void save(const std::wstring& ws)
    {
        // Prepend the string with the length in bytes, not including the terminator
        uint32_t length = ws.size() * (sizeof(wchar_t) / sizeof(char));
        save(length);

        ordered_oarchive_impl<cdr_oarchive<Order>, Order>::save(ws);
    }
};

} // namespace archive

namespace serialization {

// Provide default implementation for serializing vectors to network archives
template<marshall::byte_ordering_t Order, class T, class A>
inline void save(boost::archive::cdr_oarchive<Order> & ar, 
                 const std::vector<T, A> & v,
                 const unsigned version)
{
    // Prepend the sequence with the number of elements
    uint32_t length = v.size();
    ar << length;

    for (typename std::vector<T, A>::const_iterator it = v.begin(); it != v.end(); ++it)
        ar << (*it);
}

} // namespace serialization
} // namespace boost

#endif // CDR_OARCHIVE_HPP

