
#ifndef MARSHALL_HPP
#define MARSHALL_HPP

#include <ostream>

#include <boost/serialization/access.hpp>
#include <boost/serialization/split_member.hpp>

#include "vector_marshall.hpp"

namespace marshall {

template<class T, size_t N>
class fixed_length : public T
{
public:
    fixed_length(void) : T() { (*this).resize(N); }
    fixed_length(const T & t) : T()
    {
      if (t.size() > N)
      {
        assign(t.begin(), t.begin() + N);
      }
      else
      {
        assign(t.begin(), t.end());

        resize(N);
      }
    }

private:
    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive& ar, const unsigned version) const
    {
        if (size() <= N)
        {
            ar << static_cast<const T&>(*this);

            unsigned count = N - size();
            while (count--)
                ar << typename T::value_type();
        }
        else
        {
            T sub(begin(), begin() + N);
            ar << sub;
        }
    }

    template<class Archive>
    void load(Archive& ar, const unsigned version)
    {
        if (size() != N)
            resize(N);

        ar >> static_cast<T&>(*this);
    }

    template<class Archive>
    void serialize(Archive& ar, const unsigned version)
    {
        boost::serialization::split_member(ar, *this, version);
    }
};

template<class T, class S>
class variable_length : public T
{
public:
    variable_length(void) : T() {}
    variable_length(const T & t) : T(t) {}

private:
    friend class boost::serialization::access;

    template<class Archive>
    void save(Archive& ar, const unsigned version) const
    {
        S count = size();

        ar << count << static_cast<const T&>(*this);
    }

    template<class Archive>
    void load(Archive& ar, const unsigned version) 
    {
        S count = 0;
        ar >> count;
        
        resize(count);
        ar >> static_cast<T&>(*this);
    }

    template<class Archive>
    void serialize(Archive& ar, const unsigned version)
    {
        boost::serialization::split_member(ar, *this, version);
    }
};

} // namespace marshall

template <class T, size_t N>
inline std::ostream& operator<< (std::ostream& os, const marshall::fixed_length<T, N>& value)
{
  return os << static_cast<const T&>(value);
}

template <class T, class S>
inline std::ostream& operator<< (std::ostream& os, const marshall::variable_length<T, S>& value)
{
  return os << static_cast<const T&>(value);
}

#endif // MARSHALL_HPP

