#ifndef ORDERED_IARCHIVE_HPP
#define ORDERED_IARCHIVE_HPP

#include <istream>
#include <string>
#include <vector>

#include <boost/utility/enable_if.hpp>

#include <boost/archive/basic_binary_iprimitive.hpp>
#include <boost/serialization/split_free.hpp>

#include "byte_ordering.hpp"
#include "unidentified_archive.hpp"

#include "ordered_types.hpp"

namespace boost { 
namespace archive {

using marshall::byte_ordering_t;
using marshall::conversion_function;

// The signature for a function to load a particular size in a particular ordering
template<size_t N, byte_ordering_t Order, class Primitive>
void ordered_load(Primitive* primitive, void* address);

// A function to handle the standard load method
template<typename LoadType, typename Converter, byte_ordering_t Order, class Primitive>
inline void do_ordered_load(Primitive* primitive, void* address)
{
    static Converter converter = reinterpret_cast<Converter>(conversion_function<LoadType, Order>());

    LoadType value = 0;
    primitive->load_binary(&value, sizeof(LoadType));
    *(reinterpret_cast<LoadType*>(address)) = (*converter)(&value);
}

// The binary primitive for ordered output
template<class Archive, class IStream, byte_ordering_t Order>
struct ordered_binary_iprimitive :
    public basic_binary_iprimitive<Archive, IStream>
{
    ordered_binary_iprimitive(IStream & is, bool no_codecvt) :
        basic_binary_iprimitive<Archive, IStream>(is, no_codecvt)
    {
    }

    // Override the load function to do byte-ordered loading of the types we handle
    template<typename T>
    void load(T& t, 
              typename boost::enable_if<boost::mpl::contains<ordered_types, T>, T>::type* dummy = 0)
    {
        // When loading data, align it to required boundaries
        This()->pre_align(t);

        aligned_load<T>(t);

        This()->post_align(t);
    }

    // Ensure that a bool is encoded as 8-bit
    void load(bool & b)
    {
        uint8_t value;

        load(value);
        b = (value ? true : false);
    }

    // Assume that a string's length has been determined externally
    void load(std::string &s)
    {
        size_t size = s.size();
        char* buffer = new char[size];

        This()->pre_align(*s.begin());
        
        load_binary(buffer, size);
        s.assign(buffer, buffer + size);

        This()->post_align(s);

        delete [] buffer;
    }
    #ifndef BOOST_NO_STD_WSTRING
    void load(std::wstring &ws)
    {
        This()->pre_align(*ws.begin());

        for (std::wstring::iterator it = ws.begin(); it != ws.end(); ++it)
            aligned_load(*it);

        This()->post_align(ws);
    }
    #endif

protected:
    friend class basic_binary_iarchive<Archive>;

    // Once aligned, load data from the appropriate byte ordering
    template<typename T>
    void aligned_load(T & t)
    {
        ordered_load<sizeof(T), Order>(this, & t);
    }

    // Default alignment - no padding
    template<typename T>
    void pre_align(const T&) {}

    template<typename T>
    void post_align(const T&) {}
};

// Combine the ordered primitive with an unidentified archive
template<class Archive, byte_ordering_t Order>
struct ordered_iarchive_impl : 
    public ordered_binary_iprimitive<Archive, std::istream, Order>,
    public unidentified_iarchive<Archive>
{
    ordered_iarchive_impl(std::istream & is, unsigned int flags = no_header|no_codecvt) :
        ordered_binary_iprimitive<Archive, std::istream, Order>(
            is, 
            0 != (flags & no_codecvt))
    {
        if(0 == (flags & no_header)){
            unidentified_iarchive<Archive>::init();
            ordered_binary_iprimitive<Archive, std::istream, Order>::init();
        }
    }
};

template <byte_ordering_t Order>
struct ordered_iarchive : 
    public ordered_iarchive_impl<ordered_iarchive<Order>, Order>
{
    ordered_iarchive(std::istream & is, unsigned int flags = no_header|no_codecvt) :
        ordered_iarchive_impl<ordered_iarchive<Order>, Order>(is, flags)
    {}
};

} // namespace archive

namespace serialization {

// Provide default implementation for de-serializing vectors from network archives
template<class Archive, marshall::byte_ordering_t Order, class T, class A>
inline void load(boost::archive::ordered_iarchive_impl<Archive, Order> & ar, 
                 std::vector<T, A> & v,
                 const unsigned version)
{
    for (typename std::vector<T, A>::iterator it = v.begin(); it != v.end(); ++it)
        ar >> (*it);
}

} // namespace serialization
} // namespace boost

#endif // ORDERED_IARCHIVE_HPP
