
#ifndef TEST_HPP
#define TEST_HPP

#include <vector>

#include <boost/cstdint.hpp>
#include <boost/serialization/base_object.hpp>

#include "marshall.hpp"
#include "xdr_marshall.hpp"

// A sub-object of arbitrary complexity
struct subobject
{
    subobject(void) : a(0), b(0) {}
    subobject(uint32_t x, uint32_t y) : a(x), b(y) {}

    bool operator== (const subobject& other) const { return (a == other.a && b == other.b); }

private:
    uint32_t a;
    uint32_t b;

    friend class boost::serialization::access;

    template <class Archive>
    void serialize(Archive& archive, const unsigned version)
    {
        archive & a & b;
    }
};

// Elements common to all archive types
class test_common
{
    // All archives include these primitives
    uint32_t a;
    int32_t b;
    uint64_t c;
    int64_t d;
    bool e;
    float f;
    double g;

    friend class boost::serialization::access;

    template <class Archive>
    void serialize(Archive& archive, const unsigned version)
    {
        archive & a & b & c & d & e & f & g;
    }

public:
    test_common(void) : 
        a(0x04050607), 
        b(-0x04050607),
        c(0x08090a0b0c0d0e0f), 
        d(-0x08090a0b0c0d0e0f), 
        e(true),
        f(3.14),
        g(3.14)
    {}

    test_common(bool) {}

    bool operator== (const test_common& other) const
    {
        return (a == other.a && 
                b == other.b &&
                c == other.c &&
                d == other.d &&
                e == other.e &&
                f == other.f &&
                g == other.g);
    }
};

// Test the serializable elements of an ordered binary archive
class test_ordered : public test_common
{
    // Smaller primitives
    uint8_t h;
    int8_t i;
    uint16_t j;
    int16_t k;
    
    // String data that must be serialized to a pre-determined length
    marshall::fixed_length<std::string, 10> l;
    marshall::fixed_length<std::wstring, 8> m;

    // String data that must be serialized with a prepended length (whose size is predetermined)
    marshall::variable_length<std::string, uint8_t> n;
    marshall::variable_length<std::wstring, uint8_t> o;

    // Arbirtrary data that must be serialized to a pre-determined length
    marshall::fixed_length<std::vector<uint16_t>, 5> p;
    marshall::fixed_length<std::vector<subobject>, 2> q;

    // Arbitrary data that must be serialized with a prepended length (whose size is predetermined)
    marshall::variable_length<std::vector<uint16_t>, uint32_t> r;
    marshall::variable_length<std::vector<char>, uint32_t> s;

    friend class boost::serialization::access;

    template <class Archive>
    void serialize(Archive& archive, const unsigned version)
    {
        archive & boost::serialization::base_object<test_common>(*this) 
                & h & i & j & k & l & m & n & o & p & q & r & s;
    }

public:
    test_ordered(void) : 
        test_common(),
        h(0x01),
        i(-0x01),
        j(0x0203),
        k(-0x0203),
        l("abc"),
        n("def")
    {
        m[0] = 'A';
        m[1] = 'B';
        m[2] = 'C';

        /* Uncommenting this breaks the CDR test - yes, CDR! 
        o[0] = static_cast<wchar_t>('A');
        o[1] = static_cast<wchar_t>('B');
        o[2] = static_cast<wchar_t>('C');
        */

        p[0] = 1;
        p[1] = 2;
        p[2] = 3;

        for (uint16_t z = 1; z <= 3; ++z)
          r.push_back(10 - z);

        q[0] = subobject(99, 100);

        s.push_back('m');
    }

    test_ordered(bool initialise) : test_common(false) {}

    bool operator== (const test_ordered& other) const
    {
        return (test_common::operator==(other) &&
                h == other.h &&
                i == other.i &&
                j == other.j &&
                k == other.k &&
                l == other.l &&
                m == other.m &&
                n == other.n &&
                o == other.o &&
                p == other.p &&
                q == other.q &&
                r == other.r &&
                s == other.s);
    }
};

class test_xdr : public test_common
{
    // String data that must be serialized to a pre-determined length
    xdr_marshall::fixed_length<std::string, 10> h;

    // String data that must be serialized with a prepended 32-bit length
    std::string i;

    // Arbirtrary data that must be serialized to a pre-determined length
    xdr_marshall::fixed_length<std::vector<uint32_t>, 5> j;
    xdr_marshall::fixed_length<std::vector<subobject>, 2> k;

    // Arbitrary data that must be serialized with a prepended 32-bit length
    std::vector<int32_t> l;
    std::vector<char> m;

    friend class boost::serialization::access;

    template <class Archive>
    void serialize(Archive& archive, const unsigned version)
    {
        archive & boost::serialization::base_object<test_common>(*this) 
                & h & i & j & k & l & m;
    }

public:
    test_xdr(void) : 
        test_common(),
        h("abc"),
        i("def")
    {
        j[0] = 1;
        j[1] = 2;
        j[2] = 3;

        for (uint16_t z = 1; z <= 3; ++z)
          l.push_back(-z);

        k[0] = subobject(99, 100);

        m.push_back('m');
    }

    test_xdr(bool initialise) :
        test_common(initialise)
    {
    }

    bool operator== (const test_xdr& other) const
    {
        return (test_common::operator==(other) &&
                h == other.h &&
                i == other.i &&
                j == other.j &&
                k == other.k &&
                l == other.l &&
                m == other.m);
    }
};

class test_cdr : public test_common
{
    // Smaller primitives
    uint8_t h;
    int8_t i;
    uint16_t j;
    int16_t k;
    
    // Strings which are not wrapped in fixed- or variable-length distinguishers
    std::string l;
    std::wstring m;

    // Opaque data - sequence of bytes
    std::vector<unsigned char> n;

    friend class boost::serialization::access;

    template <class Archive>
    void serialize(Archive& archive, const unsigned version)
    {
        archive & boost::serialization::base_object<test_common>(*this) 
                & h & i & j & k & l & m & n;
    }

public:
    test_cdr(void) : 
        test_common(),
        h(0x01),
        i(-0x01),
        j(0x0203),
        k(-0x0203),
        l("abc")
    {
        m[0] = static_cast<wchar_t>('D');
        m[1] = static_cast<wchar_t>('E');
        m[2] = static_cast<wchar_t>('F');

        for (uint8_t z = 1; z <= 5; ++z)
          n.push_back(z);
    }

    test_cdr(bool initialise) :
        test_common(initialise)
    {
    }

    bool operator== (const test_cdr& other) const
    {
        return (test_common::operator==(other) &&
                h == other.h &&
                i == other.i &&
                j == other.j &&
                k == other.k &&
                l == other.l &&
                m == other.m &&
                n == other.n);
    }
};

#endif

