
#ifndef XDR_OARCHIVE_HPP
#define XDR_OARCHIVE_HPP

#include <iostream>

#include "ordered_oarchive.hpp"

#include "xdr_types.hpp"

namespace boost {
namespace archive {

using marshall::Big;

struct xdr_oarchive : 
    public ordered_oarchive_impl<xdr_oarchive, Big>
{
    xdr_oarchive(std::ostream& os) :
        ordered_oarchive_impl<xdr_oarchive, Big>(os) {}

    // Default alignment - no post-alignment
    template<typename T>
    void post_align(const T&) {}

    // Align the next element after these elements to a 4-byte boundary
    void post_align(const std::string&) { do_post_align(os.tellp()); } 

    // Sequences of opaque data must be padded out to a 4-byte boundary
    template<template<typename T> class A>
    void post_align(const std::vector<char, A<char> >&) { do_post_align(os.tellp()); } 
    template<template<typename T> class A>
    void post_align(const std::vector<unsigned char, A<unsigned char> >&) { do_post_align(os.tellp()); } 

    // Default implementation same as for ordered binary archive
    template<typename T>
    void save(const T& t, 
              typename boost::enable_if<boost::mpl::contains<xdr_types, T>, T>::type* dummy = 0)
    {
        ordered_oarchive_impl<xdr_oarchive, Big>::save(t);
    }

    // Override bool to enforce 32-bit storage
    void save(bool b)
    {
        int32_t value = (b ? 1 : 0);
        save(value);
    }

    void save(const std::string& s)
    {
        // Prepend the string with the 32-bit length
        uint32_t length = s.size();
        save(length);

        ordered_oarchive_impl<xdr_oarchive, Big>::save(s);
    }

protected:
    void do_post_align(int position)
    {
      // Ensure that the next element is on a 4-byte boundary
      if (position % 4)
      {
        static char buffer[4] = { 0 };
#ifdef INITIALISE_WITH_RECOGNISABLE_VALUE
        static bool initialised = false;

        if (initialised == false)
        {
          std::fill(buffer, buffer + sizeof(buffer), 0xda);
          initialised = true;
        }
#endif

        size_t advance_bytes = 4 - (position % 4);
        os.write(buffer, advance_bytes);
      }
    }
};

} // namespace archive

namespace serialization {

// Provide default implementation for serializing vectors to network archives
template<class T, class A>
inline void save(boost::archive::xdr_oarchive & ar, 
                 const std::vector<T, A> & v,
                 const unsigned version)
{
    // Prepend the sequence with the number of elements
    uint32_t length = v.size();
    ar << length;

    for (typename std::vector<T, A>::const_iterator it = v.begin(); it != v.end(); ++it)
        ar << (*it);

    ar.post_align(v);
}

} // namespace serialization
} // namespace boost

#endif // XDR_OARCHIVE_HPP

